import java.awt.Color;
import java.awt.Point;
import java.awt.Polygon;
import java.util.Random;
import java.util.Vector;

/////////////////////////////////////////////////////////////////////////////
class Bot extends Thread
	{
	/**
	Define a bot to be run by a separate thread.

	Public Methods:
		Bot()
		myposition()
		mycolour()
		mysize()
		getMouseInfo()
		run()
	Private Methods:
		//nextCyclicIndex()
		//lastCyclicIndex()
		moveSelf()
		getSiblings()
		avoidSiblings()

	Every bot has a position, shape, size, and colour.
	Some bots are mobile and others are not.
	(Right now, all are mobile.)

	Some mobile bots can remember a little of the past---either their
	previous positions or the mouse's previous positions, or both.
	(Right now, all do.)

	Mobile bots with memory have a position array to record
	their last three positions. They can also have velocity
	(and, eventually, acceleration).
	*/

	Soup creator;                   //my creator
	int imageWidth;                 //its dimensions
	int imageHeight;

	Random Generator;               //a pseudo-random number generator

	int mySize;                     //my size
	Color myColour;                 //my colour
	Polygon myShape;                //my shape
	Point myVelocity;               //my velocity (in x and y components)

	Point position[];               //an array of my previous locations
	//here's the new code...
	//CyclicArray position[];       //an array of my previous locations
	int bIndex;                     //index into my position array

	Point mousePosition[];          //the mouse's previous locations
	//here's the new code...
	//CyclicArray mousePosition[];  //the mouse's previous locations
	int mIndex;                     //index into mouse position array
	
	boolean allStarted = false;     //true when all my sibs have started up

	///////////////////////////////////////////////////////////////
	Bot(Soup creator, Random Generator)
		{
		/**
		This is the basic constructor for a bot.
		Initialize my position, size, shape, and colour, and the PRNG,
		then create my positions arrays, and draw myself into the bitmap.

		Globals: all variables
		*/

		//remember my creator (an applet) and get its screen dimensions
		this.creator = creator;
		imageHeight = creator.size().height;
		imageWidth = creator.size().width;

		//remember my creator's PRNG
		this.Generator = Generator;

		//generate a size for me;
		//make sure it's at least 10X10 pixels and no more than 1/10 the screen
		mySize = 20 + (int) (Generator.nextFloat() * imageHeight/20);

		//generate a starting position for me
		int x = (int) (Generator.nextFloat() * (imageWidth - mySize));
		int y = (int) (Generator.nextFloat() * (imageHeight - mySize));
		Point myPosition = new Point(x, y);

		//generate a shape for me
		int numVertices = 3 + (int) (Generator.nextFloat() * (mySize/4));
		myShape = new Polygon();
		for (int i = 0; i < numVertices; i++)
			{
			int xVertex = 1 + (int) (Generator.nextFloat() * mySize);
			int yVertex = 1 + (int) (Generator.nextFloat() * mySize);
			myShape.addPoint(xVertex, yVertex);
			}

		//generate a colour for me
		float redValue = Generator.nextFloat();
		float greenValue = Generator.nextFloat();
		float blueValue = Generator.nextFloat();
		myColour = new Color(redValue, greenValue, blueValue);

		/*
		the following code is only for mobile bots...
		it will eventually be in a separate constructor...
		*/

		//generate a (small) starting velocity for me;
		//make sure each velocity component is at least 1 pixel/cycle
		//and at most 3 pixels/cycle
		int xSpeed = 1 + (int) (Generator.nextFloat() * 3);
		int ySpeed = 1 + (int) (Generator.nextFloat() * 3);
		myVelocity = new Point(xSpeed, ySpeed);

		/*
		the following code is only for mobile bots with memory...
		this code will also eventually be in a separate constructor...
		*/

		position = new Point[3];//to remember my last three positions...
		//here's the new code...
		//position = new CyclicArray[3];
		for (int i = 0; i < 3; i++)
			position[i] = new Point(myPosition.x, myPosition.y);
		bIndex = 0;   //pointer into my positions array

		//bots with memory can remember old mouse positions...
		mousePosition = new Point[3];//to remember the mouse's last three positions
		//here's the new code...
		//mousePosition = new CyclicArray[3];
		for (int i = 0; i < 3; i++)
			mousePosition[i] = new Point(0, 0);
		mIndex = 0;   //pointer into the mouse's positions array
		}

	///////////////////////////////////////////////////////////////
	public Point botpos()
		{
		/*
		Tell the outside world my current position

		Globals: bIndex, position[]
		*/

		//this method will eventually be overridden to handle immobile bots

		return position[bIndex];
		}

	///////////////////////////////////////////////////////////////
	public int botindex()
		{
		/*
		Tell the outside world the current index into my position array

		Globals: bIndex, position[]
		*/

		//this method will eventually be overridden to handle immobile bots

		return bIndex;
		}

	///////////////////////////////////////////////////////////////
	public Point[] myposition()
		{
		/*
		Tell the outside world my position array

		Globals: bIndex, position[]
		*/

		//this method will eventually be overridden to handle immobile bots

		return position;
		}

	///////////////////////////////////////////////////////////////
	public Color mycolour()
		{
		/*
		Tell the outside world my colour

		Globals: myColour
		*/

		return myColour;
		}

	///////////////////////////////////////////////////////////////
	public Polygon myshape()
		{
		/*
		Tell the outside world my shape

		Globals: myShape
		*/

		return myShape;
		}

	///////////////////////////////////////////////////////////////
	public int mysize()
		{
		/*
		Tell the outside world my size

		Globals: mySize
		*/

		return mySize;
		}

	///////////////////////////////////////////////////////////////
	public void getMouseInfo(Point mousePos)
		{
		/**
		Get the current mouse position.
		This method is executed by my creator whenever the mouse moves.

		Globals: mIndex, mousePosition[]

		Calls: nextCyclicIndex
		*/

		//save current mouse position...
		mIndex = nextCyclicIndex(mousePosition, mIndex);
		mousePosition[mIndex] = mousePos;
		//here's the new code...
		//mIndex = mousePosition.nextCyclicIndex(mIndex);
		//mousePosition[mIndex] = mousePos;
		}

	///////////////////////////////////////////////////////////////
	private int nextCyclicIndex(Point someArray[], int someIndex)
		{
		/**
		Generate the next index after someIndex
		into a cyclic array of points someArray
		wrap the indices so that index 0 follows index someArray.length - 1.

		//note: nextCyclicIndex() and lastCyclicIndex() should become methods
		//in a separate class called ``CyclicArray''
		*/

		someIndex = someIndex + 1;
		if (someIndex > someArray.length - 1)
			someIndex = 0;
		return someIndex;
		}

	///////////////////////////////////////////////////////////////
	private int lastCyclicIndex(Point someArray[], int someIndex)
		{
		/**
		Generate the last index before someIndex
		into a cyclic array of points someArray
		wrap the indices so that index 0 follows index someArray.length - 1.
		*/

		someIndex = someIndex - 1;
		if (someIndex < 0)
			someIndex = someArray.length - 1;
		return someIndex;
		}

	/////////////////////////////////////////////////////////////////////
	private Vector getSiblings()
		{
		/**
		Find all my sibling threads and return an array of their pointers.

		This method is separate to allow for killing old bots and creating
		new bots dynamically as the world evolves.
		*/

		//find my ThreadGroup
		ThreadGroup g = this.getThreadGroup();

		//enumerate all the sibling threads in my ThreadGroup,
		//which includes my creator and any other threads my creator creates
		int numThreads = g.activeCount();
		Thread[] threads = new Thread[numThreads];
		g.enumerate(threads);

		//visit all my sibling threads,
		//skipping myself and any non-bot, null, dead, or daemon threads
		Vector siblings = new Vector();
		Thread t;      //just for shorthand in the following loop
		for (int i = 0; i < numThreads; i++)
			{
			t = threads[i];
			if ((t != null) && (t.getName().startsWith("Bot")) &&
				((t != this) && (t.isAlive())) && !(t.isDaemon()))
				{
				siblings.addElement((Bot) t);
				}
			}

		return siblings;
		}

	///////////////////////////////////////////////////////////////
	private int moveSelf(int myindex, int mouseindex)
		{
		/**
		Figure out where to move to next
		given that i'm presently at position[myindex] and the mouse is
		presently at position[mouseindex]
		return the index in position[] to move to next

		Globals: mousePosition[], position[], myVelocity,

		Calls: nextCyclicIndex(), lastCyclicIndex()
		*/

		//sense where the mouse is in relation to me
		boolean mouseBelow =
				(mousePosition[mouseindex].y > position[myindex].y);
		boolean mouseAbove =
				(mousePosition[mouseindex].y < position[myindex].y);
		boolean mouseRight =
				(mousePosition[mouseindex].x > position[myindex].x);
		boolean mouseLeft =
				(mousePosition[mouseindex].x < position[myindex].x);

		//generate a convenience pointer to the previous mouse position
		int mlast = lastCyclicIndex(mousePosition, mouseindex);
		//here's the new code...
		//int mlast = mousePosition.lastCyclicIndex(mouseindex);

		//sense whether the mouse is moving up, down, left, or right
		boolean mouseMovingDown =
				(mousePosition[mouseindex].y > mousePosition[mlast].y);
		boolean mouseMovingUp =
				(mousePosition[mouseindex].y < mousePosition[mlast].y);
		boolean mouseMovingRight =
				(mousePosition[mouseindex].x > mousePosition[mlast].x);
		boolean mouseMovingLeft =
				(mousePosition[mouseindex].x < mousePosition[mlast].x);

		//increase my speed to get away from the mouse
		if (mouseRight && mouseMovingLeft)	myVelocity.x -= 1;
		if (mouseLeft  && mouseMovingRight)	myVelocity.x += 1;
		if (mouseAbove && mouseMovingDown)	myVelocity.y += 1;
		if (mouseBelow && mouseMovingUp)	myVelocity.y -= 1;

		//generate a convenience pointer to my next position
		int bnext = nextCyclicIndex(position, myindex);
		//here's the new code...
		//int bnext = position.nextCyclicIndex(myindex);

		position[bnext].x = position[myindex].x + myVelocity.x;
		position[bnext].y = position[myindex].y + myVelocity.y;

		//now bounce off any walls i hit...
		handleCollisions(bnext);
		//return the pointer to my present location
		return bnext;
		}

	///////////////////////////////////////////////////////////////
	private int avoidSiblings(int myindex, Vector siblings)
		{
		/**
		Figure out where to move to next
		given that i'm presently at position[myindex]
		and my siblings are at (siblings[i].myposition())[myindex]
		(this method is asynchronous, so they may be at slightly different
		positions, but this is good enough for government work...)
		return the index in position[] to move to next

		Globals: position[], myVelocity

		Calls: nextCyclicIndex(), lastCyclicIndex()
		*/

		//generate a convenience pointer to my next position
		int bnext = nextCyclicIndex(position, myindex);
		//here's the new code...
		//int bnext = position.nextCyclicIndex(myindex);

		//as a first approximation, guess that i'll stay where i am
		position[bnext].x = position[myindex].x;
		position[bnext].y = position[myindex].y;

		//now see whether to perturb that...
		Point[] sibPosition = new Point[3];
		Bot sib;   //just for shorthand in the following loop
		for (int i = 0; i < siblings.size(); i++)
			{
			//for each of my siblings...

			//generate a shorthand handle for it
			sib = (Bot) siblings.elementAt(i);

			//get its current position
			Point sibpos = sib.botpos();

			//if this sib isn't close to me, ignore it
			if ((Math.abs(sibpos.x - position[myindex].x) > 2*mySize) ||
				(Math.abs(sibpos.y - position[myindex].y) > 2*mySize))
				continue;
			
			//okay, this particular sib is too close for comfort...
			//copy its position array and current index for shorthand use later
			int sibindex = sib.botindex();
			for (int j = 0; j < sibPosition.length; j++)
				sibPosition[j] = sib.myposition()[j];

			//sense where this sib is in relation to me
			boolean sibBelow =
				(sibPosition[sibindex].y > position[myindex].y);
			boolean sibAbove =
				(sibPosition[sibindex].y < position[myindex].y);
			boolean sibRight =
				(sibPosition[sibindex].x > position[myindex].x);
			boolean sibLeft =
				(sibPosition[sibindex].x < position[myindex].x);

			//generate a convenience pointer to this sib's previous position
			int mlast = lastCyclicIndex(sibPosition, sibindex);
			//here's the new code...
			//int mlast = sibPosition.lastCyclicIndex(sibindex);

			//sense whether this sib is moving up, down, left, or right
			boolean sibMovingDown =
				(sibPosition[sibindex].y > sibPosition[mlast].y);
			boolean sibMovingUp =
				(sibPosition[sibindex].y < sibPosition[mlast].y);
			boolean sibMovingRight =
				(sibPosition[sibindex].x > sibPosition[mlast].x);
			boolean sibMovingLeft =
				(sibPosition[sibindex].x < sibPosition[mlast].x);

			//increase my speed to get away from this sib
			if (sibRight && sibMovingLeft)	myVelocity.x -= 1;
			if (sibLeft  && sibMovingRight)	myVelocity.x += 1;
			if (sibAbove && sibMovingDown)	myVelocity.y += 1;
			if (sibBelow && sibMovingUp)	myVelocity.y -= 1;

			position[bnext].x += myVelocity.x;
			position[bnext].y += myVelocity.y;
			}

		//now handle any wall collisions after the above adjustments
		handleCollisions(bnext);
		return bnext;
		}

	///////////////////////////////////////////////////////////////
	private void handleCollisions(int curPos)
		{
		/**
		Bounce off any walls i hit by adjusting my myVelocity vector
		and my current position.
		I'm of size mySize and i'm stuck in a window of size
		(0..imageWidth, 0..imageHeight)

		Globals: mySize, position[], myVelocity[], imageWidth, imageHeight
		*/

		if (position[curPos].x >= imageWidth - mySize)
			{
			//i've hit a vertical wall to the right of me
			myVelocity.x = -myVelocity.x;
			position[curPos].x = 2*(imageWidth - mySize) - position[curPos].x;
			}
		if (position[curPos].x <= 0)
			{
			//i've hit a vertical wall to the left of me
			myVelocity.x = -myVelocity.x;
			position[curPos].x = -position[curPos].x;
			}
		if (position[curPos].y >= imageHeight - mySize)
			{
			//i've hit a horizontal wall below me
			myVelocity.y = -myVelocity.y;
			position[curPos].y = 2*(imageHeight - mySize) - position[curPos].y;
			}
		if (position[curPos].y <= 0)
			{
			//i've hit a horizontal wall above me
			myVelocity.y = -myVelocity.y;
			position[curPos].y = -position[curPos].y;
			}
		}

	///////////////////////////////////////////////////////////////
	public synchronized void waitToStart()
		{
		/**
		Wait until all my siblings have been started.
		The boolean allStarted will be set to true by my creator
		(by executing my startUp() method) when all my siblings
		have been started.
		I execute this method as the first thing i do; then i wait().

		Globals: allStarted
		*/

		while (!allStarted)
			{
			try {wait();}
			catch (InterruptedException ignored) {}
			}
		}

	///////////////////////////////////////////////////////////////
	public synchronized void startUp()
		{
		/**
		Start my thread up.
		This method is executed by my creator.

		Globals: allStarted
		*/

		allStarted = true;
		notifyAll();
		}

	///////////////////////////////////////////////////////////////
	public void run()
		{
		/**
		now i can live...

		Globals: bIndex, mIndex, myVelocity

		Calls: moveSelf()

		Comments:
		mIndex is a global index into the mouse position array.
		it's initialized in the constructor and updated in getMouseInfo()
		bIndex is a global index into the bot position array
		it's initialized in the constructor and updated here
		mousePosition[mIndex] is where the mouse is,
		position[bIndex] is where i am.
		*/

		int tempIndex;

		//wait until all my sibs are alive
		waitToStart();
		//enumerate them all so that i can notice their movements
		Vector siblings = new Vector();
		siblings = getSiblings();

		int timer = 0;   //to count when to apply friction and avoid my sibs
		while (true)
			{
			//figure out where to move next
			//synchronize on myself to avoid contention
			//with my creator about the mouse position
			synchronized (this)
				{
				tempIndex = moveSelf(bIndex, mIndex);
				}

			//add friction so that i slow down when left alone
			if ((timer % 10) == 0)
				{
				if (myVelocity.x > +1) myVelocity.x--;
				if (myVelocity.x < -1) myVelocity.x++;
				if (myVelocity.y > +1) myVelocity.y--;
				if (myVelocity.y < -1) myVelocity.y++;

				//reset my friction timer
				timer = 0;

				//avoid my siblings as well as the mouse
				tempIndex = avoidSiblings(tempIndex, siblings);
				}
			timer++;

			//update my position index
			bIndex = tempIndex;

			//now sleep for a bit to give any other threads a chance
			try { sleep(40); }
			catch (InterruptedException ignored) {}
			}
		}

	}//////end of class Bot///////////

/*
class CyclicArray
	{
	Point[] someArray;

	///////////////////////////////////////////////////////////////
	public int nextCyclicIndex(int someIndex)
		{
		Generate the next index after someIndex
		into a cyclic array of points someArray
		wrap the indices so that index 0 follows index someArray.length - 1.

		someIndex = someIndex + 1;
		if (someIndex > someArray.length - 1) someIndex = 0;
		return someIndex;
		}

	///////////////////////////////////////////////////////////////
	public int lastCyclicIndex(Point someArray[], int someIndex)
		{
		Generate the last index before someIndex
		into a cyclic array of points someArray
		wrap the indices so that index 0 follows index someArray.length - 1.

		someIndex = someIndex - 1;
		if (someIndex < 0) someIndex = someArray.length - 1;
		return someIndex;
		}
	}
*/
